import logbook

from regipy.exceptions import RegistryKeyNotFoundException, NoRegistryValuesException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logbook.Logger(__name__)

SERVICES_PATH = r'Services'


class ServicesPlugin(Plugin):
    NAME = 'services'
    DESCRIPTION = 'Enumerate the services in the SYSTEM hive'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.info('Started Services enumeration Plugin...')
        for control_set_services_path in self.registry_hive.get_control_sets(SERVICES_PATH):

            try:
                subkey = self.registry_hive.get_key(control_set_services_path)
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)
                continue

            self.entries[control_set_services_path] = {
                'timestamp': subkey.header.last_modified
            }
            services = []
            for service in subkey.iter_subkeys():
                values = None
                if service.values_count > 0:
                    values = [x for x in service.iter_values(as_json=True)]

                services.append({
                    'name': service.name,
                    'last_modified': convert_wintime(service.header.last_modified, as_json=self.as_json),
                    'values': values,
                    'parameters': [x for x in self.registry_hive.recurse_subkeys(nk_record=service, path=r'{}\{}'.format(
                        control_set_services_path, service.name), as_json=True)] if service.subkey_count else None
                })
                self.entries[control_set_services_path]['services'] = services



NTUSER_HIVE_TYPE = 'ntuser'
SYSTEM_HIVE_TYPE = 'system'
AMCACHE_HIVE_TYPE = 'amcache'
SOFTWARE_HIVE_TYPE = 'software'
SAM_HIVE_TYPE = 'sam'


HVLE_TRANSACTION_LOG_MAGIC = b'HvLE'
DIRT_TRANSACTION_LOG_MAGIC = b'DIRT'



import json
import os
from tempfile import mkdtemp

import pytest

from regipy.recovery import apply_transaction_logs
from regipy.regdiff import compare_hives
from regipy.registry import RegistryHive, NKRecord


def test_parse_header(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)

    assert isinstance(registry_hive, RegistryHive)
    assert registry_hive.header.primary_sequence_num == 744
    assert registry_hive.header.secondary_sequence_num == 744
    assert registry_hive.header.last_modification_time == 129782982453388850
    assert registry_hive.header.major_version == 1
    assert registry_hive.header.minor_version == 3
    assert registry_hive.header.root_key_offset == 32
    assert registry_hive.header.hive_bins_data_size == 733184
    assert registry_hive.header.minor_version == 3
    assert registry_hive.header.file_name == '?\\C:\\Users\\vibranium\\ntuser.dat'
    assert registry_hive.header.checksum == 448714443


def test_parse_root_key(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)

    assert isinstance(registry_hive, RegistryHive)
    assert isinstance(registry_hive.root, NKRecord)
    assert registry_hive.root.name == 'CMI-CreateHive{6A1C4018-979D-4291-A7DC-7AED1C75B67C}'
    assert registry_hive.root.subkey_count == 11
    assert dict(registry_hive.root.header) == {
        'access_bits': b'\x00\x00\x00\x00',
        'class_name_offset': 4294967295,
        'class_name_size': 0,
        'flags': {
            'KEY_COMP_NAME': True,
            'KEY_HIVE_ENTRY': True,
            'KEY_HIVE_EXIT': False,
            'KEY_NO_DELETE': True,
            'KEY_PREDEF_HANDLE': False,
            'KEY_SYM_LINK': False,
            'KEY_VOLATILE': False
        },
        'key_name_size': 52,
        'key_name_string': b'CMI-CreateHive{6A1C4018-979D-4291-A7DC-7AED1C75B67C}',
        'largest_sk_class_name': 0,
        'largest_sk_name': 40,
        'largest_value_name': 0,
        'last_modified': 129780243434537497,
        'largest_value_data': 0,
        'parent_key_offset': 1656,
        'security_list_offset': 1376,
        'subkey_count': 11,
        'subkeys_list_offset': 73760,
        'values_count': 0,
        'values_list_offset': 4294967295,
        'volatile_subkey_count': 2,
        'volatile_subkeys_list_offset': 2147486280
    }


def test_find_keys_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    run_key = registry_hive.get_key(r'\Software\Microsoft\Windows\CurrentVersion\Run')

    assert run_key.name == 'Run'
    assert run_key.header.last_modified == 129779615948377168

    values = [x for x in run_key.iter_values(as_json=True)]
    assert values[0].name == 'Sidebar'
    assert values[0].value_type == 'REG_EXPAND_SZ'


def test_ntuser_timeline(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    # TODO
    pass


def test_regdiff(ntuser_hive, second_hive_path):
    found_differences = compare_hives(ntuser_hive, second_hive_path)
    assert len(found_differences) == 2
    assert len([x for x in found_differences if x[0] == 'new_subkey']) == 1
    assert len([x for x in found_differences if x[0] == 'new_value']) == 1


def test_ntuser_emojis(transaction_ntuser):
    # There are some cases where the Registry stores utf-16 emojis as subkey names :)
    registry_hive = RegistryHive(transaction_ntuser)
    international = registry_hive.get_key(r'\Control Panel\International')
    subkeys = [x.name for x in international.iter_subkeys()]
    assert subkeys == ['Geo', 'User Profile', 'User Profile System Backup', '🌎🌏🌍']


def test_recurse_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    EXPECTED_KEYS = ['path', 'subkey_name', 'timestamp', 'values', 'values_count']

    value_types = {
        'REG_BINARY': 0,
        'REG_DWORD': 0,
        'REG_EXPAND_SZ': 0,
        'REG_MULTI_SZ': 0,
        'REG_NONE': 0,
        'REG_QWORD': 0,
        'REG_SZ': 0
    }

    subkey_count = 0
    values_count = 0
    for subkey in registry_hive.recurse_subkeys(as_json=True):
        subkey_values = subkey.values
        subkey_count += 1
        values_count += len(subkey_values or [])
        if subkey_values:
            for x in subkey_values:
                value_types[x['value_type']] += 1

    assert subkey_count == 2318
    assert values_count == 4613
    assert value_types == {
        'REG_BINARY': 620,
        'REG_DWORD': 1516,
        'REG_EXPAND_SZ': 100,
        'REG_MULTI_SZ': 309,
        'REG_NONE': 141,
        'REG_QWORD': 57,
        'REG_SZ': 1870
    }


def test_recurse_amcache(amcache_hive):
    registry_hive = RegistryHive(amcache_hive)

    value_types = {
        'REG_BINARY': 0,
        'REG_DWORD': 0,
        'REG_EXPAND_SZ': 0,
        'REG_MULTI_SZ': 0,
        'REG_NONE': 0,
        'REG_QWORD': 0,
        'REG_SZ': 0
    }
    subkey_count = 0
    values_count = 0
    for subkey in registry_hive.recurse_subkeys():
        subkey_count += 1
        subkey_values = subkey.values
        values_count += len(subkey_values or [])
        if subkey_values:
            for x in subkey_values:
                value_types[x.value_type] += 1
    assert subkey_count == 2118
    assert values_count == 17546
    assert value_types == {
        'REG_BINARY': 56,
        'REG_DWORD': 1656,
        'REG_EXPAND_SZ': 0,
        'REG_MULTI_SZ': 140,
        'REG_NONE': 0,
        'REG_QWORD': 1255,
        'REG_SZ': 14439
    }


def test_ntuser_apply_transaction_logs(transaction_ntuser, transaction_log):
    output_path = os.path.join(mkdtemp(), 'recovered_hive.dat')
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(transaction_ntuser, transaction_log,
                                                                             restored_hive_path=output_path)
    assert recovered_dirty_pages_count == 132

    found_differences = compare_hives(transaction_ntuser, restored_hive_path)
    assert len(found_differences) == 587
    assert len([x for x in found_differences if x[0] == 'new_subkey']) == 527
    assert len([x for x in found_differences if x[0] == 'new_value']) == 59


def test_system_apply_transaction_logs(transaction_system, system_tr_log_1, system_tr_log_2):
    output_path = os.path.join(mkdtemp(), 'recovered_hive.dat')
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(transaction_system,
                                                                             primary_log_path=system_tr_log_1,
                                                                             secondary_log_path=system_tr_log_2,
                                                                             restored_hive_path=output_path)
    assert recovered_dirty_pages_count == 315

    found_differences = compare_hives(transaction_system, restored_hive_path)
    assert len(found_differences) == 2486
    assert len([x for x in found_differences if x[0] == 'new_subkey']) == 2472
    assert len([x for x in found_differences if x[0] == 'new_value']) == 13


def test_system_apply_transaction_logs_2(transaction_usrclass, usrclass_tr_log_1, usrclass_tr_log_2):
    output_path = os.path.join(mkdtemp(), 'recovered_hive.dat')
    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(transaction_usrclass,
                                                                             primary_log_path=usrclass_tr_log_1,
                                                                             secondary_log_path=usrclass_tr_log_2,
                                                                             restored_hive_path=output_path)
    assert recovered_dirty_pages_count == 158

    found_differences = compare_hives(transaction_usrclass, restored_hive_path)
    assert len(found_differences) == 73
    assert len([x for x in found_differences if x[0] == 'new_subkey']) == 33
    assert len([x for x in found_differences if x[0] == 'new_value']) == 40


def test_hive_serialization(ntuser_hive, temp_output_file):
    registry_hive = RegistryHive(ntuser_hive)
    registry_hive.dump_hive_to_json(temp_output_file, registry_hive.root, verbose=False)
    counter = 0
    with open(temp_output_file, 'r') as dumped_hive:
        for x in dumped_hive.readlines():
            assert json.loads(x)
            counter += 1
    assert counter == 2318



import pytest

from regipy.plugins import NTUserPersistencePlugin, UserAssistPlugin, AmCachePlugin, WordWheelQueryPlugin, \
    UACStatusPlugin, LastLogonPlugin, SoftwareClassesInstallerPlugin, InstalledSoftwarePlugin, RASTracingPlugin
from regipy.plugins.ntuser.typed_urls import TypedUrlsPlugin
from regipy.plugins.software.profilelist import ProfileListPlugin
from regipy.plugins.software.persistence import SoftwarePersistencePlugin
from regipy.plugins.system.computer_name import ComputerNamePlugin
from regipy.plugins.system.shimcache import ShimCachePlugin
from regipy.registry import RegistryHive


def test_shimcache_plugin(system_hive):
    registry_hive = RegistryHive(system_hive)
    plugin_instance = ShimCachePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 660
    assert plugin_instance.entries[0] == {
        'last_mod_date': '2011-01-12T12:08:00+00:00',
        'path': '\\??\\C:\\Program Files\\McAfee\\VirusScan Enterprise\\mfeann.exe',
        'exec_flag': 'True'
    }


def test_computer_name_plugin(system_hive):
    registry_hive = RegistryHive(system_hive)
    plugin_instance = ComputerNamePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == [
        {'name': 'WKS-WIN732BITA',
         'timestamp': '2010-11-10T17:18:08.718750+00:00'
         },
        {'name': 'WIN-V5T3CSP8U4H',
         'timestamp': '2010-11-10T18:17:36.968750+00:00'
         }
    ]


def test_persistence_plugin_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    plugin_instance = NTUserPersistencePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == {
        '\\Software\\Microsoft\\Windows\\CurrentVersion\\Run': {
            'timestamp': '2012-04-03T21:19:54.837716+00:00',
            'values': [
                {'name': 'Sidebar',
                 'value_type': 'REG_EXPAND_SZ',
                 'value': '%ProgramFiles%\\Windows Sidebar\\Sidebar.exe /autoRun', 'is_corrupted': False
                 }
            ]
        }
    }


def test_persistence_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = SoftwarePersistencePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == {
        '\\Microsoft\\Windows\\CurrentVersion\\Run':
            {'timestamp': '2012-04-04T01:54:23.669836+00:00',
             'values': [
                 {
                     'name': 'VMware Tools',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\VMware\\VMware Tools\\VMwareTray.exe"',
                     'is_corrupted': False
                 },
                 {
                     'name': 'VMware User Process',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\VMware\\VMware Tools\\VMwareUser.exe"',
                     'is_corrupted': False
                 },
                 {
                     'name': 'Adobe ARM',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\Common Files\\Adobe\\ARM\\1.0\\AdobeARM.exe"',
                     'is_corrupted': False
                 },
                 {
                     'name': 'McAfeeUpdaterUI',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\McAfee\\Common Framework\\udaterui.exe" /StartedFromRunKey',
                     'is_corrupted': False
                 },
                 {
                     'name': 'ShStatEXE',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\McAfee\\VirusScan Enterprise\\SHSTAT.EXE" /STANDALONE',
                     'is_corrupted': False
                 },
                 {
                     'name': 'McAfee Host Intrusion Prevention Tray',
                     'value_type': 'REG_SZ',
                     'value': '"C:\\Program Files\\McAfee\\Host Intrusion Prevention\\FireTray.exe"',
                     'is_corrupted': False
                 },
                 {
                     'name': 'svchost',
                     'value_type': 'REG_SZ',
                     'value': 'c:\\windows\\system32\\dllhost\\svchost.exe',
                     'is_corrupted': False
                 }
             ]
             }
    }


def test_user_assist_plugin_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    plugin_instance = UserAssistPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 62
    assert plugin_instance.entries[-1] == {
        'focus_count': 1,
        'name': '%PROGRAMFILES(X86)%\\Microsoft Office\\Office14\\EXCEL.EXE',
        'run_counter': 4,
        'session_id': 0,
        'timestamp': '2012-04-04T15:43:14.785000+00:00',
        'total_focus_time_ms': 47673
    }

    assert plugin_instance.entries[50] == {
        'focus_count': 9,
        'name': 'Microsoft.Windows.RemoteDesktop',
        'run_counter': 8,
        'session_id': 0,
        'timestamp': '2012-04-03T22:06:58.124282+00:00',
        'total_focus_time_ms': 180000
    }


def test_plugin_amcache(amcache_hive):
    registry_hive = RegistryHive(amcache_hive)
    plugin_instance = AmCachePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 1120
    assert plugin_instance.entries[100] == {
        'full_path': 'C:\\Windows\\system32\\TPVMMondeu.dll',
        'last_modified_timestamp_2': '2017-03-17T05:06:04.002722+00:00',
        'program_id': '75a010066bb612ca7357ce31df8e9f0300000904',
        'sha1': '056f4b9d9ec9b5dc548e1b460da889e44089d76f',
        'timestamp': '2017-08-03T11:34:02.263418+00:00',
        'type': 'win_8+_amcache'
    }


def test_word_wheel_query_plugin_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    plugin_instance = WordWheelQueryPlugin(registry_hive, ntuser_hive)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 6
    assert plugin_instance.entries[0] == {
        'last_write': '2012-04-04T15:45:18.551340+00:00',
        'mru_id': 1,
        'name': 'alloy',
        'order': 0
    }


def test_uac_status_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = UACStatusPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == {
        'consent_prompt_admin': 5,
        'consent_prompt_user': 3,
        'enable_limited_user_accounts': 1,
        'enable_virtualization': 1,
        'filter_admin_token': 0,
        'last_write': '2011-08-30T18:47:10.734144+00:00'
    }


def test_classes_installer_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = SoftwareClassesInstallerPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries[0] == {
        'identifier': '000041091A0090400000000000F01FEC',
        'is_hidden': False,
        'product_name': 'Microsoft Office OneNote MUI (English) 2010',
        'timestamp': '2010-11-10T10:31:06.573040+00:00'
    }

    assert not any([x['is_hidden'] for x in plugin_instance.entries])


def test_ras_tracing_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = RASTracingPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 140

    assert plugin_instance.entries[0] == {
        'name': 'AcroRd32_RASAPI32',
        'timestamp': '2012-03-16T21:31:26.613878+00:00'
    }

    assert plugin_instance.entries[-1] == {
        'name': 'wmplayer_RASMANCS',
        'timestamp': '2012-03-12T20:58:55.476336+00:00'
    }


def test_installed_programs_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = InstalledSoftwarePlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert len(plugin_instance.entries) == 134

    assert plugin_instance.entries[0] == {
        'registry_path': '\\Microsoft\\Windows\\CurrentVersion\\Uninstall',
        'service_name': 'AddressBook',
        'timestamp': '2009-07-14T04:41:12.758808+00:00'
    }

    assert plugin_instance.entries[-1].items() > {
        'service_name': '{FE2F6A2C-196E-4210-9C04-2B1BC21F07EF}',
        'timestamp': '2011-07-05T22:58:57.996094+00:00',
        'registry_path': 'WOW6432Node\\Microsoft\\Windows\\CurrentVersion\\Uninstall',
        'UninstallString': 'MsiExec.exe /X{FE2F6A2C-196E-4210-9C04-2B1BC21F07EF}',
        'URLInfoAbout': 'http://www.vmware.com',
        'DisplayName': 'VMware Tools'
    }.items()


def test_last_logon_plugin_software(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = LastLogonPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == {
        'last_logged_on_provider': '{6F45DC1E-5384-457A-BC13-2CD81B0D28ED}',
        'last_logged_on_sam_user': 'SHIELDBASE\\rsydow',
        'last_logged_on_user': 'SHIELDBASE\\rsydow',
        'last_write': '2012-04-04T12:20:41.453654+00:00',
        'show_tablet_keyboard': 0
    }


def test_typed_urls_plugin_ntuser(ntuser_hive):
    registry_hive = RegistryHive(ntuser_hive)
    plugin_instance = TypedUrlsPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == {
        'last_write': '2012-04-03T22:37:55.411500+00:00',
        'entries': [
            {'url1': 'http://199.73.28.114:53/'},
            {'url2': 'http://go.microsoft.com/fwlink/?LinkId=69157'}
        ]
    }


def test_profilelist_plugin(software_hive):
    registry_hive = RegistryHive(software_hive)
    plugin_instance = ProfileListPlugin(registry_hive, as_json=True)
    plugin_instance.run()

    assert plugin_instance.entries == [{
        "last_write": "2009-07-14T04:41:12.493608+00:00",
        "path": "%systemroot%\\system32\\config\\systemprofile",
        "flags": 12,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-18",
        "load_time": None,
        "local_load_time": None
    }, {
        "last_write": "2010-11-10T18:09:16.250000+00:00",
        "path": "C:\\Windows\\ServiceProfiles\\LocalService",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-19",
        "load_time": None,
        "local_load_time": None
    }, {
        "last_write": "2010-11-10T18:09:16.250000+00:00",
        "path": "C:\\Windows\\ServiceProfiles\\NetworkService",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-20",
        "load_time": None,
        "local_load_time": None
    }, {
        "last_write": "2010-11-10T17:22:52.109376+00:00",
        "path": "C:\\Users\\Pepper",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-100689374-1717798114-2601648136-1000",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2012-04-04T12:42:17.719834+00:00",
        "path": "C:\\Users\\SRL-Helpdesk",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-100689374-1717798114-2601648136-1001",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2011-08-21T00:51:19.820166+00:00",
        "path": "C:\\Users\\nfury",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1105",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2011-08-23T01:33:29.006350+00:00",
        "path": "C:\\Users\\mhill",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1106",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2011-09-17T13:33:17.372366+00:00",
        "path": "C:\\Users\\Tdungan",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1107",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2012-04-06T19:44:17.844274+00:00",
        "path": "C:\\Users\\nromanoff",
        "flags": 0,
        "full_profile": None,
        "state": 0,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1109",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2012-04-06T19:42:31.408714+00:00",
        "path": "C:\\Users\\rsydow",
        "flags": 0,
        "full_profile": None,
        "state": 256,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1114",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }, {
        "last_write": "2012-04-06T19:22:20.845938+00:00",
        "path": "C:\\Users\\vibranium",
        "flags": 0,
        "full_profile": None,
        "state": 256,
        "sid": "S-1-5-21-2036804247-3058324640-2116585241-1673",
        "load_time": "1601-01-01T00:00:00+00:00",
        "local_load_time": None
    }
    ]
